import decimal
import json
from datetime import datetime, date


class DateTimeEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, (datetime, date)):
            return o.isoformat()
        if isinstance(o, decimal.Decimal):
            return float(o)
        return super().default(o)


def dumps(o):
    return json.dumps(o, cls=DateTimeEncoder)



import logging
import logging.handlers
import sys
from pythonjsonlogger import jsonlogger
from watchmen_boot.config.config import PROD, settings


def init():
    logger = logging.getLogger()
    logger.setLevel(logging.NOTSET)

    # Add stdout handler, with level INFO
    console = logging.StreamHandler(sys.stdout)
    if settings.ENVIRONMENT == PROD:
        console.setLevel(logging.ERROR)
    else:
        console.setLevel(logging.INFO)
    formatter = logging.Formatter('%(name)-13s: %(levelname)-8s %(message)s')
    console.setFormatter(formatter)
    logger.addHandler(console)
    # Add file rotating handler, with level DEBUG
    if settings.LOGGER_FILE_ON:
        if settings.LOGGER_JSON_FORMAT:
            log_handler = logging.handlers.RotatingFileHandler(filename=settings.LOGGER_FILE, maxBytes=10242880,
                                                               backupCount=5, encoding='utf-8')
            log_handler.setLevel(logging.INFO)
            formatter = jsonlogger.JsonFormatter(
                '%(asctime)s - %(process)d - %(threadName)s - %(name)s - %(levelname)s - %(message)s')
            log_handler.setFormatter(formatter)
            logger.addHandler(log_handler)
        else:
            rotating_handler = logging.handlers.RotatingFileHandler(filename='temp/rotating.log', maxBytes=10242880,
                                                                    backupCount=5, encoding='utf-8')
            rotating_handler.setLevel(logging.ERROR)
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            rotating_handler.setFormatter(formatter)
            logger.addHandler(rotating_handler)



from datetime import datetime


class MongoStorage:

    def __init__(self, engine):
        self.engine = engine
        self.worker_id_table = "snowflake_workerid"

    def insert_one(self, record):
        collection = self.engine.get_collection( self.worker_id_table)
        collection.insert_one(self.__convert_to_dict(record))

    def update_one(self, ip, process_id):
        collections = self.engine.get_collection(self.worker_id_table)
        where_ = {"$and": [{"ip": {"$eq": ip}}, {"processid": {"$eq": process_id}}]}
        update_ = {"$set": {"regdate": datetime.now()}}
        collections.update_one(where_, update_)

    def find_(self, model, ip, process_id):
        collection = self.engine.get_collection(self.worker_id_table)
        where_ = {"$and": [{"ip": {"$eq": ip}}, {"processid": {"$eq": process_id}}]}
        cursor = collection.find(where_)
        result_list = list(cursor)
        return [model.parse_obj(result) for result in result_list]

    def list_all(self, model):
        collection = self.engine.get_collection(self.worker_id_table)
        cursor = collection.find()
        result_list = list(cursor)
        return [model.parse_obj(result) for result in result_list]

    def delete_by_id(self, ip_, process_id):
        collection = self.engine.get_collection(self.worker_id_table)
        where_ = {"$and": [{"ip": {"$eq": ip_}}, {"processid": {"$eq": process_id}}]}
        collection.delete_one(where_)

    def __convert_to_dict(self, instance) -> dict:
        if type(instance) is not dict:
            return instance.dict(by_alias=True)
        else:
            return instance



import requests

from watchmen_boot.config.config import settings

session = requests.Session()
session.headers.update({"Content-Type": "application/json"})


def next_id():
    url = settings.SNOWFLAKE_REMOTE_HOST + "/snowflakeid"
    response = session.get(url)
    if response.status_code == 200:
        return response.text



from typing import List
from pydantic import BaseModel

from watchmen_boot.model.watchmen_model import WatchmenModel


class DataSourceParam(BaseModel):
    name: str = None
    value: str = None


class DataSource(WatchmenModel):
    dataSourceId: str = None
    dataSourceCode: str = None
    dataSourceType: str = None
    host: str = None
    port: str = None
    username: str = None
    password: str = None
    name: str = None
    url: str = None
    params: List[DataSourceParam] = []
    tenantId: str = None



from datetime import datetime
from operator import eq

from sqlalchemy import MetaData, Table, Column, String, Integer, Date, insert, update, and_, select, delete

from watchmen_boot.storage.oracle.oracle_utils import parse_obj


class OracleStorage:

    def __init__(self, engine):
        self.engine = engine
        self.metadata = MetaData()
        self.worker_id_table = Table("snowflake_workerid", self.metadata,
                                     Column('ip', String(100), primary_key=True),
                                     Column('processid', String(60), primary_key=True),
                                     Column('workerid', Integer, nullable=False),
                                     Column('regdate', Date, nullable=True)
                                     )

    def insert_one(self, record):
        table = self.worker_id_table
        stmt = insert(table).values(record)
        with self.engine.connect() as conn:
            with conn.begin():
                conn.execute(stmt)

    def update_one(self, ip, process_id):
        table = self.worker_id_table
        stmt = update(table)
        filter_ = [eq(table.c["ip"], ip), eq(table.c["processid"], process_id)]
        stmt = stmt.where(and_(*filter_))
        stmt = stmt.values({"regdate": datetime.now()})
        with self.engine.connect() as conn:
            with conn.begin():
                conn.execute(stmt)

    def find_(self, model, ip, process_id):
        table = self.worker_id_table
        filter_ = [eq(table.c["ip"], ip), eq(table.c["processid"], process_id)]
        stmt = select(table).where(and_(*filter_))
        with self.engine.connect() as conn:
            cursor = conn.execute(stmt).cursor
            columns = [col[0] for col in cursor.description]
            cursor.rowfactory = lambda *args: dict(zip(columns, args))
            row = cursor.fetchone()
            if row is None:
                return None
            else:
                return parse_obj(model, row, table)

    def list_all(self, model):
        table = self.worker_id_table
        stmt = select(table)
        with self.engine.connect() as conn:
            cursor = conn.execute(stmt).cursor
            columns = [col[0] for col in cursor.description]
            cursor.rowfactory = lambda *args: dict(zip(columns, args))
            res = cursor.fetchall()
        results = []
        for row in res:
            results.append(parse_obj(model, row, table))
        return results

    def delete_by_id(self, ip_, process_id):
        table = self.worker_id_table
        filter_ = [eq(table.c["ip"], ip_), eq(table.c["processid"], process_id)]
        stmt = delete(table).where(and_(*filter_))
        with self.engine.connect() as conn:
            with conn.begin():
                conn.execute(stmt)



from datetime import datetime, date

from bson import ObjectId
from pydantic import BaseModel, BaseConfig


class WatchmenModel(BaseModel):
    lastModified: datetime = datetime.now().replace(tzinfo=None)
    createTime: str = None

    class Config(BaseConfig):
        json_encoders = {
            datetime: lambda dt: dt.isoformat(),
            ObjectId: lambda oid: str(oid),
            date: lambda dt: dt.isoformat()
        }



from watchmen_boot.config.config import settings
from watchmen_boot.constants.database.constant import MONGO, MYSQL, ORACLE
from watchmen_boot.utils.datasource_utils import get_default_datasource


def find_template():
    default_datasource = get_default_datasource()
    if settings.STORAGE_ENGINE == MONGO:
        from watchmen_boot.storage.mongo.mongo_client import MongoEngine
        from watchmen_boot.guid.storage.mongo.mongo_template import MongoStorage
        engine = MongoEngine(default_datasource)
        return MongoStorage(engine.get_engine())
    elif settings.STORAGE_ENGINE == MYSQL:
        from watchmen_boot.storage.mysql.mysql_client import MysqlEngine
        from watchmen_boot.guid.storage.mysql.mysql_template import MysqlStorage
        engine = MysqlEngine(default_datasource)
        return MysqlStorage(engine.get_engine())
    elif settings.STORAGE_ENGINE == ORACLE:
        from watchmen_boot.storage.oracle.oracle_client import OracleEngine
        from watchmen_boot.guid.storage.oracle.oracle_template import OracleStorage
        engine = OracleEngine(default_datasource)
        return OracleStorage(engine.get_engine())


import arrow
from math import ceil

DATETIME = "datetime"  # YYYY-MM-DD HH:mm:ss
FULL_DATETIME = "full-datetime"  # YYYY-MM-DD HH:mm:ss.SSS
DATE = "date"  # YYYY - MM - DD
TIME = "time"  # HH: mm:ss
YEAR = "year"  # 4
HALF_YEAR = "half-year"  # 1: first half, 2: second half

QUARTER = "quarter"  # 1 - 4
SEASON = "season"  # 1: spring, 2: summer, 3: autumn, 4: winter
MONTH = "month"  # 1 - 12
HALF_MONTH = "half-month"  # 1: first half, 2: second half

TEN_DAYS = "ten-days"  # 1, 2, 3
WEEK_OF_YEAR = "week-of-year"  # 0 - 53
WEEK_OF_MONTH = "week-of-month"  # 0 - 5
HALF_WEEK = "half-week"  # 1: first half, 2: second half

DAY_OF_MONTH = "day-of-month"  # 1 - 31, according to month/year

DAY_OF_WEEK = "day-of-week"  # 1(Sunday) - 7(Saturday)
DAY_KIND = "day-kind"  # 1: workday, 2: weekend, 3: holiday
HOUR = "hour"  # 0 - 23
HOUR_KIND = "hour-kind"  # 1: work time, 2: off hours, 3: sleeping time

MINUTE = "minute"  # 0 - 59
SECOND = "second"  # 0 - 59
MILLISECOND = "millisecond"  # 0-999
AM_PM = "am-pm"  # 1, 2


def parsing_and_formatting(value, format_type_):
    if value is not None:
        arr_value = arrow.get(value).datetime.replace(tzinfo=None)
        if format_type_ == YEAR:
            return arr_value.year
        elif format_type_ == HALF_YEAR:
            month = arr_value.month
            if month <= 6:
                return 1
            else:
                return 2
        elif format_type_ == QUARTER:
            return quarter_of(arr_value)
        elif format_type_ == MONTH:
            return arr_value.month
        elif format_type_ == WEEK_OF_YEAR:
            return int(arr_value.strftime("%U"))
        elif format_type_ == WEEK_OF_MONTH:
            return week_of_month(arr_value.date())
        elif format_type_ == DAY_OF_MONTH:
            return arr_value.day
        elif format_type_ == DAY_OF_WEEK:
            return int(arr_value.strftime("%w")) + 1
    else:
        return value


def quarter_of(dt):
    month_ = dt.strftime("%m")
    if month_ in ['01', '02', '03']:
        return 1
    elif month_ in ['04', '05', '06']:
        return 2
    elif month_ in ['07', '08', '09']:
        return 3
    elif month_ in ['10', '11', '12']:
        return 4


def week_of_month(dt):
    """ Returns the week of the month for the specified date.
    """
    first_day = dt.replace(day=1)
    first_day_weekday = int(first_day.strftime(
        "%w"))  # The position of the first day in a week, is also the number of days to be supplemented in that week
    days = dt.day  # the number of days about dt
    adjusted_ = first_day_weekday + days
    print(adjusted_)
    return int(ceil(adjusted_ / 7.0)) - 1



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['watchmen_boot',
 'watchmen_boot.cache',
 'watchmen_boot.config',
 'watchmen_boot.constants',
 'watchmen_boot.constants.database',
 'watchmen_boot.guid',
 'watchmen_boot.guid.storage',
 'watchmen_boot.guid.storage.mongo',
 'watchmen_boot.guid.storage.mysql',
 'watchmen_boot.guid.storage.oracle',
 'watchmen_boot.logging',
 'watchmen_boot.model',
 'watchmen_boot.storage',
 'watchmen_boot.storage.model',
 'watchmen_boot.storage.mongo',
 'watchmen_boot.storage.mysql',
 'watchmen_boot.storage.oracle',
 'watchmen_boot.storage.utility',
 'watchmen_boot.utils']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.27,<2.0.0',
 'arrow>=1.1.0,<2.0.0',
 'cacheout>=0.13.1,<0.14.0',
 'cx-Oracle>=8.2.1,<9.0.0',
 'mysqlclient>=2.1.0,<3.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'pymongo>=3.11.4,<4.0.0',
 'python-dotenv>=0.19.2,<0.20.0',
 'python-json-logger>=2.0.2,<3.0.0',
 'requests>=2.25.1,<3.0.0']

setup_kwargs = {
    'name': 'watchmen-boot',
    'version': '15.2.2',
    'description': '',
    'long_description': None,
    'author': 'luke0623',
    'author_email': 'luke0623@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from sqlalchemy import create_engine

from watchmen_boot.storage.model.data_source import DataSource
from watchmen_boot.storage.utility.date_utils import dumps


class MysqlEngine(object):

    def __init__(self, datasource: DataSource):
        self.connection_url = 'mysql://%s:%s@%s:%s/%s?charset=utf8' % (datasource.username,
                                                                       datasource.password,
                                                                       datasource.host,
                                                                       datasource.port,
                                                                       datasource.name)
        self.engine = create_engine(self.connection_url,
                                    echo=False,
                                    future=True,
                                    pool_recycle=3600,
                                    json_serializer=dumps, encoding='utf-8')

    def get_engine(self):
        return self.engine



# coding: utf-8
import logging
import os
import random
import threading
import time
from datetime import datetime
from pydantic import BaseModel

import socket

from watchmen_boot.config.config import settings
from watchmen_boot.guid.remoteid import next_id
from watchmen_boot.guid.storage.choise_storage import find_template

logger = logging.getLogger(__name__)


def get_host_ip():
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    finally:
        s.close()
        return ip


class InvalidSystemClock(Exception):
    pass


class SnowFlakeIdWorker(object):
    TWEPOCH = 1420041600000
    WORKER_ID_BITS = 10
    DATACENTER_ID_BITS = 2
    SEQUENCE_BITS = 10
    MAX_WORKER_ID = -1 ^ (-1 << WORKER_ID_BITS)  # 2**10-1 0b1111111111
    MAX_DATACENTER_ID = -1 ^ (-1 << DATACENTER_ID_BITS)
    MAX_SEQUENCE = -1 ^ (-1 << SEQUENCE_BITS)
    WOKER_ID_SHIFT = SEQUENCE_BITS
    DATACENTER_ID_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS
    TIMESTAMP_LEFT_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS + DATACENTER_ID_BITS

    def __init__(self, datacenter_id=0, sequence=0):
        gen = WorkerIdGen()
        worker_id = gen.generate_worker_id()
        threading.Thread(target=WorkerIdGen.heart_beat, args=(gen,), daemon=True).start()
        # sanity check
        if worker_id > self.MAX_WORKER_ID or worker_id < 0:
            raise ValueError('worker_id max value')

        if datacenter_id > self.MAX_DATACENTER_ID or datacenter_id < 0:
            raise ValueError('datacenter_id max value')

        self.worker_id = worker_id
        self.datacenter_id = datacenter_id
        self.sequence = sequence

        self.last_timestamp = -1  #

    def _gen_timestamp(self):
        """
        generate a timestamp
        :return:int timestamp
        """
        return int(time.time() * 1000)

    def get_id(self):
        """
        get a new id
        :return:
        """
        timestamp = self._gen_timestamp()

        #
        if timestamp < self.last_timestamp:
            raise InvalidSystemClock

        if timestamp == self.last_timestamp:
            self.sequence = (self.sequence + 1) & self.MAX_SEQUENCE
            if self.sequence == 0:
                timestamp = self._til_next_millis(self.last_timestamp)
        else:
            self.sequence = 0

        self.last_timestamp = timestamp

        new_id = ((timestamp - self.TWEPOCH) << self.TIMESTAMP_LEFT_SHIFT) | (
                self.datacenter_id << self.DATACENTER_ID_SHIFT) | \
                 (self.worker_id << self.WOKER_ID_SHIFT) | self.sequence
        return new_id

    def _til_next_millis(self, last_timestamp):
        """
        next timestamp id
        """
        timestamp = self._gen_timestamp()
        while timestamp <= last_timestamp:
            timestamp = self._gen_timestamp()
        return timestamp


class WorkerIdGen:

    def __init__(self):
        self.ip = get_host_ip()
        self.process_id = str(os.getpid())
        self.storage_template = find_template()

    def generate_worker_id(self):
        workers = self.storage_template.list_all(WorkerId)
        result = self.check_worker_status(workers)
        if result:
            return result.workerId
        else:
            new_worker_id = self.random_worker_id(workers)
            return new_worker_id

    def check_worker_status(self, workers):
        result = None
        registered_workers = workers
        for el in registered_workers:
            if (datetime.now().replace(microsecond=0) - el.regDate).days >= 1:
                self.storage_template.delete_by_id(el.ip, el.processId)
                workers.remove(el)
            else:
                if el.ip == self.ip and el.processId == self.process_id:
                    result = el
                    break
        return result

    def random_worker_id(self, workers):
        new_worker_id = random.randrange(0, 1024)
        result = self.check_worker_id(workers, new_worker_id)
        if result:
            self.storage_template.insert_one(
                {"ip": self.ip, "processid": self.process_id, "workerid": new_worker_id,
                 "regdate": datetime.now().replace(microsecond=0)})
            return new_worker_id
        else:
            return self.random_worker_id(workers)

    def check_worker_id(self, workers, new_worker_id):
        registered_workers = workers
        for el in registered_workers:
            if el.workerId == new_worker_id:
                return False
        return True

    def heart_beat(self):
        try:
            while True:
                self.storage_template.update_one(self.ip, self.process_id)
                time.sleep(30)
        finally:
            logger.error("worker id register heart beat stop")


class WorkerId(BaseModel):
    ip: str = None
    processId: str = None
    workerId: int = None
    regDate: datetime = None


worker = SnowFlakeIdWorker(settings.SNOWFLAKE_DATACENTER)


def get_surrogate_key():
    return str(get_int_surrogate_key())


def get_int_surrogate_key():
    if settings.SNOWFLAKE_REMOTE:
        return next_id()
    else:
        return worker.get_id()



from typing import List, Dict, Any, Optional

from pydantic import AnyHttpUrl, BaseSettings, validator

DEV = "dev"
PROD = "production"


class Settings(BaseSettings):
    API_V1_STR: str = ""
    # SECRET_KEY: str = secrets.token_urlsafe(32)
    SECRET_KEY: str = '801GtEAdlE8o-iZRLBMgz30PGE_zxry82EaUYMAhNq8'
    ACCESS_TOKEN_EXPIRE_MINUTES: int = 60 * 24 * 8
    BACKEND_CORS_ORIGINS: List[AnyHttpUrl] = []

    LOGGER_FILE: str = "temp/rotating.log"
    LOGGER_JSON_FORMAT: bool = False
    LOGGER_FILE_ON: bool = True

    DASK_ON: bool = False
    DASK_PROCESSES: bool = False
    DASK_TEMP: str = None

    ENVIRONMENT: str = DEV



    ALGORITHM = "HS256"
    STORAGE_ENGINE = "mongo"
    PROJECT_NAME: str
    MONGO_HOST: str = None
    MONGO_PORT: int = None
    MONGO_DATABASE: str = "watchmen"
    MONGO_USERNAME: str = None
    MONGO_PASSWORD: str = None
    PRESTO_HOST: str = None
    PRESTO_PORT: int = None
    PRESTO_USER = "the_user"
    PRESTO_CATALOG = "mongo"
    PRESTO_SCHEMA = "watchmen"
    PRESTO_ON = True
    PRESTO_LIB = "trino"

    DEFAULT_DATA_ZONE_ON = False

    MYSQL_HOST: str = ""
    MYSQL_PORT: int = 3306
    MYSQL_USER: str = ""
    MYSQL_PASSWORD: str = ""
    MYSQL_DATABASE: str = "watchmen"
    MYSQL_POOL_MAXCONNECTIONS: int = 6
    MYSQL_POOL_MINCACHED = 2
    MYSQL_POOL_MAXCACHED = 5
    MYSQL_ECHO = False

    ORACLE_LIB_DIR: str = ""
    ORACLE_HOST: str = ""
    ORACLE_PORT: int = 1521
    ORACLE_USER: str = ""
    ORACLE_PASSWORD: str = ""
    ORACLE_SERVICE: str = ""
    ORACLE_SID: str = ""
    ORACLE_NAME: str = ""

    CONNECTOR_KAFKA = False
    CONNECTOR_RABBITMQ = False

    RABBITMQ_HOST: str = ""
    RABBITMQ_PORT: str = "5672"
    RABBITMQ_USERNAME: str = ""
    RABBITMQ_PASSWORD: str = ""
    RABBITMQ_VIRTUALHOST: str = ""
    RABBITMQ_QUEUE: str = ""
    RABBITMQ_DURABLE: bool = True
    RABBITMQ_AUTO_DELETE: bool = False

    KAFKA_BOOTSTRAP_SERVER = "localhost:9092"
    KAFKA_TOPICS = ""

    NOTIFIER_PROVIDER = "email"
    EMAILS_ENABLED: bool = False
    SMTP_TLS: bool = True
    SMTP_PORT: Optional[int] = None
    SMTP_HOST: Optional[str] = None
    SMTP_USER: Optional[str] = None
    SMTP_PASSWORD: Optional[str] = None
    EMAILS_FROM_EMAIL: Optional[str] = None
    EMAILS_FROM_NAME: Optional[str] = None
    EMAILS_TO: Optional[str] = None
    TOPIC_DATE_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
    DECIMAL = "decimal(32,2)"

    MOCK_USER = "demo_user"
    SNOWFLAKE_DATACENTER = 0
    SNOWFLAKE_WORKER = 0
    SNOWFLAKE_REMOTE = False
    SNOWFLAKE_REMOTE_HOST: str = None

    MULTIPLE_DATA_SOURCE = False
    EXTERNAL_WRITER_ON = True

    PROMETHEUS_ON = False

    INDEX_ON = False

    PIPELINE_MONITOR_ON = True

    QUERY_MONITOR_ON = False

    # 'aws-kms' 'azure-kms' , 'aliyun-kms'
    KEY_MANAGEMENT_TYPE: str = "db"

    DATA_SECURITY_ON = False

    @validator("STORAGE_ENGINE", pre=True)
    def check_storage_configuration(cls, v: str, values: Dict[str, Any]) -> bool:
        # print(v)
        if v and v == "mongo":
            result = bool(
                values.get("MONGO_HOST")
                and values.get("MONGO_PORT"))
            if not result:
                raise ValueError("STORAGE_ENGINE dependency check MONGO_HOST and MONGO_PORT")
            else:
                return v
        elif v == "mysql":
            result = bool(
                values.get("MYSQL_HOST")
                and values.get("MYSQL_PORT")
                and values.get("MYSQL_USER")
            )
            if not result:
                raise ValueError("STORAGE_ENGINE dependency check MYSQL_HOST and MYSQL_PORT and MYSQL_USER")
            else:
                return v
        elif v == "oracle":
            result = bool(
                values.get("ORACLE_HOST")
                and values.get("ORACLE_PORT")
                and values.get("ORACLE_USER")
                and values.get("ORACLE_LIB_DIR")
                and values.get("ORACLE_SERVICE")
            )
            if not result:
                raise ValueError("STORAGE_ENGINE dependency check ORACLE_HOST and ORACLE_PORT and ORACLE_USER")
            else:
                return v

    class Config:
        env_file = '.env'
        env_file_encoding = 'utf-8'
        case_sensitive = True


settings = Settings()
